@@ -1,5 +1,5 @@
 use crate::config::dns_canister_config::DnsCanisterConfig;
-use clap::{crate_authors, crate_version, AppSettings, Clap};
+use clap::{crate_authors, crate_version, AppSettings, Parser};
 use hyper::{
     body,
     body::Bytes,
@@ -8,14 +8,21 @@ use hyper::{
     Body, Client, Request, Response, Server, StatusCode, Uri,
 };
 use ic_agent::{
-    agent::http_transport::ReqwestHttpReplicaV2Transport, export::Principal, Agent, AgentError,
+    agent::http_transport::ReqwestHttpReplicaV2Transport,
+    export::Principal,
+    ic_types::{hash_tree::LookupResult, HashTree},
+    lookup_value, Agent, AgentError, Certificate,
 };
 use ic_utils::{
+    call::AsyncCall,
     call::SyncCall,
     interfaces::http_request::{
-        HeaderField, HttpRequestCanister, StreamingCallbackHttpResponse, StreamingStrategy,
+        HeaderField, HttpRequestCanister, HttpResponse, StreamingCallbackHttpResponse,
+        StreamingStrategy,
     },
 };
+use lazy_regex::regex_captures;
+use sha2::{Digest, Sha256};
 use slog::Drain;
 use std::{
     convert::Infallible,
@@ -38,12 +45,11 @@ static MAX_HTTP_REQUEST_STREAM_CALLBACK_CALL_COUNT: i32 = 1000;
 // The maximum length of a body we should log as tracing.
 static MAX_LOG_BODY_SIZE: usize = 100;
 
-#[derive(Clap)]
+#[derive(Parser)]
 #[clap(
     version = crate_version!(),
     author = crate_authors!(),
-    global_setting = AppSettings::GlobalVersion,
-    global_setting = AppSettings::ColoredHelp
+    global_setting = AppSettings::PropagateVersion,
 )]
 pub(crate) struct Opts {
     /// Verbose level. By default, INFO will be used. Add a single `-v` to upgrade to
@@ -84,6 +90,11 @@ pub(crate) struct Opts {
     #[clap(long)]
     debug: bool,
 
+    /// Whether or not to fetch the root key from the replica back end. Do not use this when
+    /// talking to the Internet Computer blockchain mainnet as it is unsecure.
+    #[clap(long)]
+    fetch_root_key: bool,
+
     /// A map of domain names to canister IDs.
     /// Format: domain.name:canister-id
     #[clap(long)]
@@ -196,17 +207,19 @@ async fn forward_request(
         .inspect(|HeaderField(name, value)| {
             slog::trace!(logger, "<< {}: {}", name, value);
         })
-        .collect();
+        .collect::<Vec<_>>();
 
     let entire_body = body::to_bytes(request.into_body()).await?.to_vec();
 
     slog::trace!(logger, "<<");
     if logger.is_trace_enabled() {
-        let body = String::from_utf8_lossy(&entire_body);
+        let body = String::from_utf8_lossy(
+            &entire_body[0..usize::min(entire_body.len(), MAX_LOG_BODY_SIZE)],
+        );
         slog::trace!(
             logger,
-            "<< {}{}",
-            &body[0..usize::min(body.len(), MAX_LOG_BODY_SIZE)],
+            "<< \"{}\"{}",
+            &body.escape_default(),
             if body.len() > MAX_LOG_BODY_SIZE {
                 format!("... {} bytes total", body.len())
             } else {
@@ -216,29 +229,117 @@ async fn forward_request(
     }
 
     let canister = HttpRequestCanister::create(agent.as_ref(), canister_id);
-    let result = canister
-        .http_request(method, uri.to_string(), headers, &entire_body)
+    let query_result = canister
+        .http_request(
+            method.clone(),
+            uri.to_string(),
+            headers.clone(),
+            &entire_body,
+        )
         .call()
         .await;
 
-    // If the result is a Replica error, returns the 500 code and message. There is no information
-    // leak here because a user could use `dfx` to get the same reply.
-    let (http_response,) = match result {
-        Ok(response) => response,
-        Err(AgentError::ReplicaError {
-            reject_code,
-            reject_message,
-        }) => {
-            return Ok(Response::builder()
+    fn handle_result(
+        result: Result<(HttpResponse,), AgentError>,
+    ) -> Result<HttpResponse, Result<Response<Body>, Box<dyn Error>>> {
+        // If the result is a Replica error, returns the 500 code and message. There is no information
+        // leak here because a user could use `dfx` to get the same reply.
+        match result {
+            Ok((http_response,)) => Ok(http_response),
+            Err(AgentError::ReplicaError {
+                reject_code,
+                reject_message,
+            }) => Err(Ok(Response::builder()
                 .status(StatusCode::INTERNAL_SERVER_ERROR)
                 .body(format!(r#"Replica Error ({}): "{}""#, reject_code, reject_message).into())
-                .unwrap());
+                .unwrap())),
+            Err(e) => Err(Err(e.into())),
         }
-        Err(e) => return Err(e.into()),
+    }
+
+    let http_response = match handle_result(query_result) {
+        Ok(http_response) => http_response,
+        Err(response_or_error) => return response_or_error,
     };
 
+    let http_response = if http_response.upgrade == Some(true) {
+        let waiter = garcon::Delay::builder()
+            .throttle(std::time::Duration::from_millis(500))
+            .timeout(std::time::Duration::from_secs(15))
+            .build();
+        let update_result = canister
+            .http_request_update(method, uri.to_string(), headers, &entire_body)
+            .call_and_wait(waiter)
+            .await;
+        let http_response = match handle_result(update_result) {
+            Ok(http_response) => http_response,
+            Err(response_or_error) => return response_or_error,
+        };
+        http_response
+    } else {
+        http_response
+    };
+
+    let mut certificate: Option<Result<Vec<u8>, ()>> = None;
+    let mut tree: Option<Result<Vec<u8>, ()>> = None;
+
     let mut builder = Response::builder().status(StatusCode::from_u16(http_response.status_code)?);
     for HeaderField(name, value) in http_response.headers {
+        if name.eq_ignore_ascii_case("IC-CERTIFICATE") {
+            for field in value.split(',') {
+                if let Some((_, name, b64_value)) = regex_captures!("^(.*)=:(.*):$", field.trim()) {
+                    slog::trace!(logger, ">> certificate {}: {}", name, b64_value);
+                    let bytes = base64::decode(b64_value).map_err(|e| {
+                        slog::warn!(
+                            logger,
+                            "Unable to decode {} in ic-certificate from base64: {}",
+                            name,
+                            e
+                        );
+                    });
+                    if name == "certificate" {
+                        certificate = Some(match (certificate, bytes) {
+                            (None, bytes) => bytes,
+                            (Some(Ok(certificate)), Ok(bytes)) => {
+                                slog::warn!(logger, "duplicate certificate field: {:?}", bytes);
+                                Ok(certificate)
+                            }
+                            (Some(Ok(certificate)), Err(_)) => {
+                                slog::warn!(
+                                    logger,
+                                    "duplicate certificate field (failed to decode)"
+                                );
+                                Ok(certificate)
+                            }
+                            (Some(Err(_)), bytes) => {
+                                slog::warn!(
+                                    logger,
+                                    "duplicate certificate field (failed to decode)"
+                                );
+                                bytes
+                            }
+                        });
+                    } else if name == "tree" {
+                        tree = Some(match (tree, bytes) {
+                            (None, bytes) => bytes,
+                            (Some(Ok(tree)), Ok(bytes)) => {
+                                slog::warn!(logger, "duplicate tree field: {:?}", bytes);
+                                Ok(tree)
+                            }
+                            (Some(Ok(tree)), Err(_)) => {
+                                slog::warn!(logger, "duplicate tree field (failed to decode)");
+                                Ok(tree)
+                            }
+                            (Some(Err(_)), bytes) => {
+                                slog::warn!(logger, "duplicate tree field (failed to decode)");
+                                bytes
+                            }
+                        });
+                    }
+                }
+            }
+        }
+
         builder = builder.header(&name, value);
     }
 
@@ -299,6 +400,34 @@ async fn forward_request(
 
         builder.body(body)?
     } else {
+        let body_valid = match (certificate, tree) {
+            (Some(Ok(certificate)), Some(Ok(tree))) => match validate_body(
+                &certificate,
+                &tree,
+                &canister_id,
+                &agent,
+                &uri,
+                &http_response.body,
+                logger.clone(),
+            ) {
+                Ok(valid) => valid,
+                Err(e) => {
+                    return Ok(Response::builder()
+                        .status(StatusCode::INTERNAL_SERVER_ERROR)
+                        .body(format!("Certificate validation failed: {}", e).into())
+                        .unwrap());
+                }
+            },
+            (Some(_), _) | (_, Some(_)) => false,
+            // Canisters don't have to provide certified variables
+            (None, None) => true,
+        };
+        if !body_valid {
+            return Ok(Response::builder()
+                .status(StatusCode::INTERNAL_SERVER_ERROR)
+                .body("Body does not pass verification".into())
+                .unwrap());
+        }
         builder.body(http_response.body.into())?
     };
 
@@ -321,14 +450,9 @@ async fn forward_request(
         slog::trace!(logger, ">>");
         slog::trace!(
             logger,
-            ">> {}{}",
-            match std::str::from_utf8(&body) {
-                Ok(s) => format!(
-                    r#""{}""#,
-                    s[..usize::min(MAX_LOG_BODY_SIZE, s.len())].escape_default()
-                ),
-                Err(_) => hex::encode(&body[..usize::min(MAX_LOG_BODY_SIZE, body.len())]),
-            },
+            ">> \"{}\"{}",
+            String::from_utf8_lossy(&body[..usize::min(MAX_LOG_BODY_SIZE, body.len())])
+                .escape_default(),
             if is_streaming {
                 "... streaming".to_string()
             } else if body.len() > MAX_LOG_BODY_SIZE {
@@ -342,6 +466,76 @@ async fn forward_request(
     Ok(response)
 }
 
+fn validate_body(
+    certificate: &[u8],
+    tree: &[u8],
+    canister_id: &Principal,
+    agent: &Agent,
+    uri: &Uri,
+    response_body: &[u8],
+    logger: slog::Logger,
+) -> anyhow::Result<bool> {
+    let cert: Certificate =
+        serde_cbor::from_slice(certificate).map_err(AgentError::InvalidCborData)?;
+    let tree: HashTree = serde_cbor::from_slice(tree).map_err(AgentError::InvalidCborData)?;
+
+    if let Err(e) = agent.verify(&cert) {
+        slog::trace!(logger, ">> certificate failed verification: {}", e);
+        return Ok(false);
+    }
+
+    let certified_data_path = vec![
+        "canister".into(),
+        canister_id.into(),
+        "certified_data".into(),
+    ];
+    let witness = match lookup_value(&cert, certified_data_path) {
+        Ok(witness) => witness,
+        Err(e) => {
+            slog::trace!(
+                logger,
+                ">> Could not find certified data for this canister in the certificate: {}",
+                e
+            );
+            return Ok(false);
+        }
+    };
+    let digest = tree.digest();
+
+    if witness != digest {
+        slog::trace!(
+            logger,
+            ">> witness ({}) did not match digest ({})",
+            hex::encode(witness),
+            hex::encode(digest)
+        );
+
+        return Ok(false);
+    }
+
+    let path = ["http_assets".into(), uri.path().into()];
+    let tree_sha = match tree.lookup_path(&path) {
+        LookupResult::Found(v) => v,
+        _ => match tree.lookup_path(&["http_assets".into(), "/index.html".into()]) {
+            LookupResult::Found(v) => v,
+            _ => {
+                slog::trace!(
+                    logger,
+                    ">> Invalid Tree in the header. Does not contain path {:?}",
+                    path
+                );
+                return Ok(false);
+            }
+        },
+    };
+
+    let mut sha256 = Sha256::new();
+    sha256.update(response_body);
+    let body_sha = sha256.finalize();
+
+    Ok(&body_sha[..] == tree_sha)
+}
+
 fn is_hop_header(name: &str) -> bool {
     name.to_ascii_lowercase() == "connection"
         || name.to_ascii_lowercase() == "keep-alive"
@@ -406,7 +600,7 @@ async fn forward_api(
     request: Request<Body>,
     replica_url: &str,
 ) -> Result<Response<Body>, Box<dyn Error>> {
-    let proxied_request = create_proxied_request(ip_addr, &replica_url, request)?;
+    let proxied_request = create_proxied_request(ip_addr, replica_url, request)?;
 
     let client = Client::builder().build(hyper_tls::HttpsConnector::new());
     let response = client.request(proxied_request).await?;
@@ -419,6 +613,13 @@ fn not_found() -> Result<Response<Body>, Box<dyn Error>> {
         .body("Not found".into())?)
 }
 
+fn unable_to_fetch_root_key() -> Result<Response<Body>, Box<dyn Error>> {
+    Ok(Response::builder()
+        .status(StatusCode::INTERNAL_SERVER_ERROR)
+        .body("Unable to fetch root key".into())?)
+}
+
+#[allow(clippy::too_many_arguments)]
 async fn handle_request(
     ip_addr: IpAddr,
     request: Request<Body>,
@@ -426,6 +627,7 @@ async fn handle_request(
     proxy_url: Option<String>,
     dns_canister_config: Arc<DnsCanisterConfig>,
     logger: slog::Logger,
+    fetch_root_key: bool,
     debug: bool,
 ) -> Result<Response<Body>, Infallible> {
     let request_uri_path = request.uri().path();
@@ -459,8 +661,11 @@ async fn handle_request(
                 .build()
                 .expect("Could not create agent..."),
         );
-
-        forward_request(request, agent, dns_canister_config.as_ref(), logger.clone()).await
+        if fetch_root_key && agent.fetch_root_key().await.is_err() {
+            unable_to_fetch_root_key()
+        } else {
+            forward_request(request, agent, dns_canister_config.as_ref(), logger.clone()).await
+        }
     } {
         Err(err) => {
             slog::warn!(logger, "Internal Error during request:\n{:#?}", err);
@@ -491,6 +696,7 @@ fn main() -> Result<(), Box<dyn Error>> {
     let counter = AtomicUsize::new(0);
     let debug = opts.debug;
     let proxy_url = opts.proxy.clone();
+    let fetch_root_key = opts.fetch_root_key;
 
     let service = make_service_fn(|socket: &hyper::server::conn::AddrStream| {
         let ip_addr = socket.remote_addr();
@@ -520,6 +726,7 @@ fn main() -> Result<(), Box<dyn Error>> {
                     proxy_url.clone(),
                     dns_canister_config,
                     logger,
+                    fetch_root_key,
                     debug,
                 )
             }))
